@@ -1,11 +1,13 @@
 from __future__ import annotations as _annotations
 
 import ast
+import asyncio
 import dataclasses
 import importlib.util
 import inspect
 import re
 import sys
+from collections.abc import Sequence
 from dataclasses import dataclass
 from importlib.abc import Loader
 from pathlib import Path
@@ -14,7 +16,7 @@ from typing import TYPE_CHECKING, Any, Callable
 from unittest.mock import patch
 
 import pytest
-from black import InvalidInput
+from black.parsing import InvalidInput
 
 from .lint import black_format, code_diff
 from .traceback import create_example_traceback
@@ -23,34 +25,64 @@ if TYPE_CHECKING:
     from .config import ExamplesConfig
     from .find_examples import CodeExample
 
-__all__ = 'run_code', 'InsertPrintStatements'
+__all__ = 'run_code', 'InsertPrintStatements', 'IncludePrint'
 
 parent_frame_id = 4 if sys.version_info >= (3, 8) else 3
+IncludePrint = Callable[[Path, inspect.FrameInfo, Sequence[Any]], bool]
 
 
 def run_code(
+    *,
     example: CodeExample,
     python_file: Path,
     loader: Loader | None,
     config: ExamplesConfig,
     enable_print_mock: bool,
     print_callback: Callable[[str], str] | None,
+    include_print: IncludePrint | None,
     module_globals: dict[str, Any] | None,
+    call: str | None,
 ) -> tuple[InsertPrintStatements, dict[str, Any]]:
+    """Run the code example.
+
+    Args:
+        example: The `CodeExample` to run.
+        python_file: The path to the python file.
+        loader: optional loader to use to load the module.
+        config: The `ExamplesConfig` to use.
+        enable_print_mock: If True, mock the `print` function.
+        print_callback: If not None, a callback to call on `print`.
+        include_print: If not None, a function to call to determine if the print statement should be included.
+        module_globals: The extra globals to add before calling the module.
+        call: If not None, a (coroutine) function to call in the module.
+
+    Returns:
+        A tuple of the `InsertPrintStatements` instance and the module's globals.
+    """
     __tracebackhide__ = True
 
     spec = importlib.util.spec_from_file_location('__main__', str(python_file), loader=loader)
+    assert spec is not None, f'Could not load {python_file}'
+    assert spec.loader is not None, f'Loader is None for {python_file}'
     module = importlib.util.module_from_spec(spec)
 
     # does nothing if insert_print_statements is False
-    insert_print = InsertPrintStatements(python_file, config, enable_print_mock, print_callback)
+    insert_print = InsertPrintStatements(python_file, config, enable_print_mock, print_callback, include_print)
 
     if module_globals:
         module.__dict__.update(module_globals)
+
     try:
         with insert_print:
             sys.modules[spec.name] = module
             spec.loader.exec_module(module)
+            if call:
+                to_call = getattr(module, call, None)
+                if to_call is not None:
+                    if inspect.iscoroutinefunction(to_call):
+                        asyncio.run(to_call())
+                    else:
+                        to_call()
     except KeyboardInterrupt:
         print('KeyboardInterrupt in example')
     except Exception as exc:
@@ -65,37 +97,39 @@ def run_code(
 
 @dataclass(init=False)
 class Arg:
-    string: str | None = None
-    code: str | None = None
+    """A single argument to a print statement."""
+
+    data: str
+    is_str: bool = False
 
     def __init__(self, v: Any):
         if isinstance(v, str):
-            self.string = v
+            self.data = v
+            self.is_str = True
         elif isinstance(v, set):
             # NOTE! this is not recursive
             ordered = ', '.join(repr(x) for x in sorted(v))
-            self.string = f'{{{ordered}}}'
+            self.data = f'{{{ordered}}}'
         else:
-            self.code = re.sub('0x[a-f0-9]{8,12}>', '0x0123456789ab>', str(v))
+            self.data = re.sub('0x[a-f0-9]{8,12}>', '0x0123456789ab>', str(v))
 
     def __str__(self) -> str:
-        if self.string is not None:
-            return self.string
-        else:
-            return self.code
+        return self.data
 
     def format(self, config: ExamplesConfig) -> str:
-        if self.string is not None:
-            return self.string
+        if self.is_str:
+            return self.data
         else:
             try:
-                return black_format(self.code, config)
+                return black_format(self.data, config)
             except InvalidInput:
-                return self.code
+                return self.data
 
 
 @dataclass
 class PrintStatement:
+    """A single print statement."""
+
     line_no: int
     sep: str
     args: list[Arg]
@@ -111,26 +145,51 @@ def not_print(*args):
 
 
 class MockPrintFunction:
-    def __init__(self, file: Path) -> None:
+    __slots__ = 'file', 'statements', 'include_print'
+
+    def __init__(self, file: Path, include_print: IncludePrint | None) -> None:
         self.file = file
         self.statements: list[PrintStatement] = []
+        self.include_print = include_print
 
     def __call__(self, *args: Any, sep: str = ' ', **kwargs: Any) -> None:
         frame = inspect.stack()[parent_frame_id]
 
-        if self.file.samefile(frame.filename):
-            # -1 to account for the line number being 1-indexed
-            s = PrintStatement(frame.lineno, sep, [Arg(arg) for arg in args])
+        if self._include_file(frame, args):
+            lineno = self._find_line_number(frame)
+            s = PrintStatement(lineno, sep, [Arg(arg) for arg in args])
             self.statements.append(s)
 
+    def _include_file(self, frame: inspect.FrameInfo, args: Sequence[Any]) -> bool:
+        if self.include_print:
+            return self.include_print(self.file, frame, args)
+        else:
+            return self.file.samefile(frame.filename)
+
+    def _find_line_number(self, inspect_frame: inspect.FrameInfo) -> int:
+        """Find the line number of the print statement in the file that is being executed."""
+        frame = inspect_frame.frame
+        while True:
+            if self.file.samefile(frame.f_code.co_filename):
+                return frame.f_lineno
+            elif frame.f_back:
+                frame = frame.f_back
+            else:
+                raise RuntimeError(f'Could not find line number of print statement at {inspect_frame}')
+
 
 class InsertPrintStatements:
     def __init__(
-        self, python_path: Path, config: ExamplesConfig, enable: bool, print_callback: Callable[[str], str] | None
+        self,
+        python_path: Path,
+        config: ExamplesConfig,
+        enable: bool,
+        print_callback: Callable[[str], str] | None,
+        include_print: IncludePrint | None,
     ):
         self.file = python_path
         self.config = config
-        self.print_func = MockPrintFunction(python_path) if enable else None
+        self.print_func = MockPrintFunction(python_path, include_print) if enable else None
         self.print_callback = print_callback
         self.patch = None
 
@@ -144,14 +203,16 @@ class InsertPrintStatements:
             self.patch.stop()
 
     def check_print_statements(self, example: CodeExample) -> None:
-        with_prints = self._insert_print_statements(example)
-        if example.source != with_prints:
-            diff = code_diff(example, with_prints)
+        new_code = self.updated_print_statements(example)
+        if new_code is not None:
+            diff = code_diff(example, new_code, self.config)
             pytest.fail(f'Print output changed code:\n{indent(diff, "  ")}', pytrace=False)
 
     def updated_print_statements(self, example: CodeExample) -> str | None:
         with_prints = self._insert_print_statements(example)
-        if example.source != with_prints:
+        # we check against the raw `with_prints` and `with_prints` with trailing whitespace removed
+        # since trailing white space will have already been stripped by pre-commit in `example.source`
+        if example.source not in (with_prints, re.sub(r'[ \t]+\n', '\n', with_prints)):
             return with_prints
 
     def print_statements(self) -> list[PrintStatement]:
@@ -180,9 +241,8 @@ class InsertPrintStatements:
     def _insert_print_args(
         self, lines: list[str], statement: PrintStatement, in_python: bool, line_index: int, col: int
     ) -> None:
-        single_line = statement.sep.join(map(str, statement.args))
-        if self.print_callback:
-            single_line = self.print_callback(single_line)
+        formatted_args = [self.print_callback(str(arg)) if self.print_callback else str(arg) for arg in statement.args]
+        single_line = statement.sep.join(formatted_args)
         indent_str = ' ' * col
         max_single_length = self.config.line_length - len(indent_str)
         if '\n' not in single_line and len(single_line) + len(comment_prefix) < max_single_length:
@@ -203,28 +263,11 @@ class InsertPrintStatements:
 
 comment_prefix = '#> '
 comment_prefix_re = re.compile(f'^ *{re.escape(comment_prefix)}', re.MULTILINE)
-triple_quotes_prefix_re = re.compile("^ *(?:\"{3}|'{3})", re.MULTILINE)
-
-
-def find_print_line(lines: list[str], line_no: int) -> int:
-    """
-    For 3.7 we have to reverse through lines to find the print statement lint
-    """
-    if sys.version_info >= (3, 8):
-        return line_no
-
-    for back in range(100):
-        new_line_no = line_no - back
-        m = re.search(r'^ *print\(', lines[new_line_no - 1])
-        if m:
-            return new_line_no
-    return line_no
+triple_quotes_prefix_re = re.compile('^ *(?:"{3}|\'{3})', re.MULTILINE)
 
 
 def remove_old_print(lines: list[str], line_index: int) -> None:
-    """
-    Remove the old print statement.
-    """
+    """Remove the old print statement."""
     try:
         next_line = lines[line_index + 1]
     except IndexError:
@@ -232,7 +275,7 @@ def remove_old_print(lines: list[str], line_index: int) -> None:
         return
 
     if triple_quotes_prefix_re.search(next_line):
-        for i in range(2, 100):
+        for i in range(2, 10_000):
             if triple_quotes_prefix_re.search(lines[line_index + i]):
                 del lines[line_index + 1 : line_index + i + 1]
                 return
@@ -247,24 +290,14 @@ def remove_old_print(lines: list[str], line_index: int) -> None:
 
 
 def find_print_location(example: CodeExample, line_no: int) -> tuple[int, int]:
-    """
-    Find the line and column of the print statement.
+    """Find the line and column of the print statement.
 
-    :param example: the `CodeExample`
-    :param line_no: The line number on which the print statement starts (or approx on 3.7)
-    :return: tuple if `(line, column)` of the print statement
-    """
-    # For 3.7 we have to reverse through lines to find the print statement lint
-    if sys.version_info < (3, 8):
-        # find the last print statement before the line
-        lines = example.source.splitlines()
-        for back in range(100):
-            new_line_no = line_no - back
-            m = re.match(r' *print\(', lines[new_line_no - 1])
-            if m:
-                line_no = new_line_no
-                break
+    Args:
+        example: the `CodeExample`
+        line_no: The line number on which the print statement starts or approx
 
+    Return: tuple if `(line, column)` of the print statement
+    """
     m = ast.parse(example.source, filename=example.path.name)
     return find_print(m, line_no) or (line_no, 0)
 
